/*
 * This file is part of City Domination, which is a GameScript for OpenTTD
 * Copyright (C) 2011  Daniel Beal
 *
 * City Domination is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * City Domination is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with City Domination; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

import("util.superlib","SuperLib",19);
Helper <- SuperLib.Helper;
Tile <- SuperLib.Tile;
Log <- SuperLib.Log;

class Goal
{
	id = null;
	company = null;
	town = null;
	time = null;
	constructor(company, town, id = null, time = null) {
		if(!id) this.id = GSGoal.New(company, GSText(GSText.STR_GOAL, town), GSGoal.GT_TOWN, 0);
		else this.id = id;
		this.company = company;
		this.town = town;
		if(!time) this.time = 0;
		else this.time = time;
		GSNews.Create(GSNews.NT_GENERAL, GSText(GSText.STR_NEW_GOAL, town), company);
	}
	function ExportTable();
	function VerifyGoal();
}

function Goal::ExportTable()
{
	return {
		id = this.id,
		company = this.company,
		town = this.town,
		time = this.time
	};
}

function Goal::VerifyGoal()
{
	return GSGoal.IsValidGoal(id);
}

class Town
{
	id = null;
	owner = null;
	constructor(id, owner) {
		this.id = id;
		this.owner = owner;
	}
	function ChangeOwner(newOwner);
}

function Town::ChangeOwner(newOwner)
{
	owner = newOwner;
	GSTown.ExpandTown(id, 5);
	Log.Info(GSCompany.GetName(newOwner) + " took over " + GSTown.GetName(id), Log.LVL_SUB_DECISIONS);
	GSNews.Create(GSNews.NT_ECONOMY,GSText(GSText.STR_TOWN_TAKEN_OVER, id, newOwner),GSCompany.COMPANY_INVALID);
}

class CityDomination extends GSController
{
	lastYear = null;
	newCompanies = null;
	towns = null;
	goals = null;
	constructor() {
		newCompanies = GSList();
		towns = [];
		goals = [];
	}
	function Start();
	function Save();
	function Load(version, data);
	function CheckEvents();
	function CheckMap();
	function CheckGoals();
	function CreateGoal(company);
	function UpdateTownOwners(town = null);
}

function CityDomination::Start()
{
	local townlist = GSTownList();
	foreach(town, _ in townlist) towns.append(Town(town, GSCompany.COMPANY_INVALID));
	UpdateTownOwners();
	lastYear = GSDate.GetYear(GSDate.GetCurrentDate());
	Log.Info("City Domination Started!", Log.LVL_SUB_DECISIONS);
	GSController.Sleep(1);
	while(true) {
		CheckEvents();
		CheckMap();
		CheckGoals();
		GSController.Sleep(10);
	}
}

function CityDomination::CheckEvents()
{
	while(GSEventController.IsEventWaiting()) {
		local event = GSEventController.GetNextEvent();
		if(event == null) continue;
		local eventType = event.GetEventType();
		if(eventType == GSEvent.ET_COMPANY_BANKRUPT) {
			// If the company owned towns, set them to no owner
			Log.Info("Found bankrupt company!", Log.LVL_SUB_DECISIONS);
			local deadcompany = GSEventCompanyBankrupt.Convert(event);
			foreach(town in towns) {
				if(town.owner == deadcompany.GetCompanyID()) town.owner == GSCompany.COMPANY_INVALID;
			}
		}
		if(eventType == GSEvent.ET_COMPANY_NEW) {
			Log.Info("Found new company!", Log.LVL_SUB_DECISIONS);
			local newcompany = GSEventCompanyNew.Convert(event);
			GSNews.Create(GSNews.NT_GENERAL, GSText(GSText.STR_PLACE_HQ), newcompany.GetCompanyID());
			newCompanies.AddItem(newcompany.GetCompanyID(),0);
		}
	}
}

function CityDomination::CheckMap()
{
	foreach(company, _ in newCompanies) {
		if(GSCompany.GetCompanyHQ(company) != GSMap.TILE_INVALID) {
			Log.Info("Company built HQ: " + GSCompany.GetName(company), Log.LVL_SUB_DECISIONS);
			Helper.SetSign(GSCompany.GetCompanyHQ(company),GSText(GSText.STR_DOMINATED_TOWNS, 0), true);
			GSNews.Create(GSNews.NT_GENERAL, GSText(GSText.STR_NEW_HQ, company, GSTile.GetClosestTown(GSCompany.GetCompanyHQ(company))), GSCompany.COMPANY_INVALID);
			for(local i = 0;i < GSController.GetSetting("goals_per_company");i++) CreateGoal(company);
			newCompanies.RemoveItem(company);
		}
	}
	local companies = GSList();
	for(local competitor = GSCompany.COMPANY_FIRST;competitor <= GSCompany.COMPANY_LAST;competitor++) {
		if(GSCompany.ResolveCompanyID(competitor) != GSCompany.COMPANY_INVALID) companies.AddItem(competitor,0);
	}
	foreach(town in towns) {
		// Get a list of all stations inside the town
		local tilelist = Tile.GetTownTiles(town.id);
		tilelist.Valuate(GSTile.IsStationTile);
		tilelist.KeepValue(1);
		local stationlist = GSList();
		foreach(tile, _ in tilelist) {
			if(!stationlist.HasItem(GSStation.GetStationID(tile))) {
				Log.Info("Adding station " + GSStation.GetName(GSStation.GetStationID(tile)));
				stationlist.AddItem(GSStation.GetStationID(tile), GSStation.GetCargoRating(GSStation.GetStationID(tile), Helper.GetPAXCargo()));
			}
		}
		local companylist = companies;
		if(stationlist.IsEmpty()) {
//			Log.Info("Town " + GSTown.GetName(town) + " has no stations in it");
			continue;
		}
		// Now get a list of companies get a score for each of them
		local companylist = GSList();
		for(local competitor = GSCompany.COMPANY_FIRST;competitor <= GSCompany.COMPANY_LAST;competitor++) {
			if(GSCompany.ResolveCompanyID(competitor) != GSCompany.COMPANY_INVALID) companylist.AddItem(competitor,0);
		}
		foreach(station, score in stationlist) {
			companylist.SetValue(GSStation.GetOwner(station), companylist.GetValue(GSStation.GetOwner(station)) + score);
		}
		// Finally, determine the company that should own the city
		companylist.Sort(GSList.SORT_BY_VALUE, false);
		local oldowner = town.owner;
		Log.Info("Old Owner: " + GSCompany.GetName(oldowner));
		if(oldowner != companylist.Begin()) town.ChangeOwner(companylist.Begin());
		Log.Info("New Owner: " + GSCompany.GetName(companylist.Begin()));
		if(oldowner != town.owner) {
			UpdateTownOwners(town);
		}
		Log.Info("Winning Company: " + GSCompany.GetName(town.owner));
		companies.SetValue(town.owner, companies.GetValue(town.owner) + 1);
		GSController.Sleep(1);
	}
	// Replace the HQ Domination numbers
	foreach(company, score in companies) Helper.SetSign(GSCompany.GetCompanyHQ(company), GSText(GSText.STR_DOMINATED_TOWNS, score), true);
	// If its a new year, share the results
	if(lastYear < GSDate.GetYear(GSDate.GetCurrentDate()) && !companies.IsEmpty()) {
		local bestcompany = null;
		local bestscore = 0;
		local tie = false;
		foreach(company, score in companies) {
			if(score > bestscore) {
				bestscore = score;
				bestcompany = company;
				tie = false;
			}
			else if(score == bestscore) {
				tie = true;
			}
		}
		if(!tie) GSNews.Create(GSNews.NT_GENERAL, GSText(GSText.STR_YEAR_RESULTS, bestcompany, bestscore), GSCompany.COMPANY_INVALID);
		else GSNews.Create(GSNews.NT_GENERAL, GSText(GSText.STR_YEAR_TIE, bestscore), GSCompany.COMPANY_INVALID);
		lastYear = GSDate.GetYear(GSDate.GetCurrentDate());
	}
}

function CityDomination::CheckGoals()
{
	Log.Info("Checking goals...");
	for(local i = 0;i < goals.len();i++) {
		local gtown = null;
		foreach(town in towns) {
			if(goals[i].town == town.id) {
				gtown = town;
				break;
			}
		}
		if(!gtown) {
			Log.Warning("Could not find town for goal!");
			continue;
		}
		if(!GSTown.IsValidTown(gtown.id)) {
			Log.Warning("Could not find town for goal: " + GSTown.GetName(town.goal));
			continue;
		}
		// Find out the company has taken the town
		if(goals[i].time == 0 && goals[i].company == gtown.owner) {
			// Start the clock!
			Log.Info("Started clock for goal in company " + GSCompany.GetName(goals[i].company) + " in town " + GSTown.GetName(goals[i].town));
			goals[i].time = GSDate.GetCurrentDate();
		}
		// Find out if they were not able to keep the town long enough
		else if(goals[i].time != 0 && goals[i].company != gtown.owner) {
			// Reset the clock
			Log.Info("Reset clock for goal in company " + GSCompany.GetName(goals[i].company) + " in town " + GSTown.GetName(goals[i].town));
			goals[i].time = 0;
		}
		// Find out if they were able to keep the town to meet the needs of the goal
		else if(goals[i].time != 0 && goals[i].time < GSDate.GetCurrentDate() - 90) {
			// Do some good stuff :)
			Log.Info(GSCompany.GetName(goals[i].company) + " completed their goal in " + GSTown.GetName(goals[i].town), Log.LVL_SUB_DECISIONS);
			GSNews.Create(GSNews.NT_GENERAL, GSText(GSText.STR_COMPLETED_GOAL, goals[i].town), goals[i].company);
			GSTown.ExpandTown(gtown.id, 20);
			// Create a new goal
			CreateGoal(goals[i].company);
			// Remove the old goal
			GSGoal.Remove(goals[i].id);
			goals.remove(i);
			i--;
		}
	}
}

function CityDomination::CreateGoal(company)
{
	Log.Info("Creating new goal for company " + GSCompany.GetName(company), Log.LVL_SUB_DECISIONS);
	// Put various towns as a factor for choosing their next town to occupy:
	local townlist = GSTownList();
	// (currently, only) The Comany HQ (so that the company grows in a circle around its hq
	// Remove the towns they already own
	foreach(town in towns) {
		if(company == town.owner) townlist.RemoveItem(town.id);
	}
	if(townlist.IsEmpty()) {
		Log.Warning("Either the company " + GSCompany.GetName(company) + " has taken over the entire map, or there is something wrong with the goalfinding.");
		return;
	}
	townlist.Valuate(GSTown.GetDistanceManhattanToTile, GSCompany.GetCompanyHQ(company));
	townlist.Sort(GSList.SORT_BY_VALUE, true);
	local chosenTown = townlist.Begin();
	// Ensure no duplicate goals
	foreach(goal in goals) {
		while(company == goal.company && chosenTown == goal.town) chosenTown = townlist.Next();
	}
	// Finish by creating the goal
	goals.append(Goal(company, chosenTown));
	if(!goals[goals.len()-1].VerifyGoal) {
		Log.Warning("Could not create goal: " + GSError.GetLastErrorString());
		goals.remove(goals.len()-1);
	}
}

function CityDomination::UpdateTownOwners(town = null)
{
	if(!town) {
		Log.Info("Updating town owners for each town...");
		foreach(atown in towns) {
			if(GSCompany.ResolveCompanyID(atown.owner) == GSCompany.COMPANY_INVALID) Helper.SetSign(GSTown.GetLocation(atown.id), GSText(GSText.STR_NOT_OWNED), true);
			else signid = Helper.SetSign(GSTown.GetLocation(atown.id), GSText(GSText.STR_OWNER, GSCompany.GetName(atown.owner)), true);
			// GSSign.BuildSign(GSTown.GetLocation(atown.id), Text(GSText.STR_NOT_OWNED));
			GSController.Sleep(1);
		}
	}
	else {
		Log.Info("Updating town owners for town " + GSTown.GetName(town.id), Log.LVL_SUB_DECISIONS);
		if(GSCompany.ResolveCompanyID(town.owner) == GSCompany.COMPANY_INVALID) Helper.SetSign(GSTown.GetLocation(GSText(GSText.STR_NOT_OWNED)), true);
		else Helper.SetSign(GSTown.GetLocation(town.id), GSText(GSText.STR_OWNER, town.owner), true);
		GSController.Sleep(1);
	}
}

function CityDomination::Save()
{
	Log.Info("Saving game...");
	local data = {};
	foreach(goal in goals) data[goal.id] <- goal.ExportTable();
	return data;
}

function CityDomination::Load(version, data)
{
	Log.Info("Loading from save: Version " + version);
	if(version == 1) {
		Log.Warning("Version 1 did not support save/load, your goals will be reset (even though they are still there)");
		return;
	}
	foreach(goal in data) goals.append(Goal(goal.company, goal.town, goal.id, goal.time));
}
